import numpy as np
from sklearn.feature_selection.univariate_selection import _BaseFilter, f_classif, _clean_nans
from sklearn.utils.validation import check_is_fitted


class SelectThreshold(_BaseFilter):
    """Select features according to the threshold.

    Read more in the :ref:`User Guide <univariate_feature_selection>`.

    Parameters
    ----------
    score_func : callable
        Function taking two arrays X and y, and returning a pair of arrays
        (scores, pvalues) or a single array with scores.
        Default is f_classif (see below "See also"). The default function only
        works with classification tasks.

    thresh : int or "none", optional, default=0.1
        Score threshold for feature inclusion.
        The "none" option bypasses selection, for use in a parameter search.

    Attributes
    ----------
    scores_ : array-like, shape=(n_features,)
        Scores of features.

    pvalues_ : array-like, shape=(n_features,)
        p-values of feature scores, None if `score_func` returned only scores.

    Notes
    -----
    Ties between features with equal scores will be broken in an unspecified
    way.

    See also
    --------
    f_classif: ANOVA F-value between label/feature for classification tasks.
    mutual_info_classif: Mutual information for a discrete target.
    chi2: Chi-squared stats of non-negative features for classification tasks.
    f_regression: F-value between label/feature for regression tasks.
    mutual_info_regression: Mutual information for a continuous target.
    SelectPercentile: Select features based on percentile of the highest scores.
    SelectFpr: Select features based on a false positive rate test.
    SelectFdr: Select features based on an estimated false discovery rate.
    SelectFwe: Select features based on family-wise error rate.
    GenericUnivariateSelect: Univariate feature selector with configurable mode.
    """

    def __init__(self, score_func=f_classif, thresh=0.1):
        super(SelectThreshold, self).__init__(score_func)
        self.thresh = thresh

    def _check_params(self, X, y):
        if not (self.thresh == "none" or 0 <= self.thresh <= 1):
            raise ValueError("thresh should be >=0, <= 1; got %r."
                             "Use thresh='none' to return all features."
                             % self.thresh)

    def _get_support_mask(self):
        check_is_fitted(self, 'scores_')

        if self.thresh == 'none':
            return np.ones(self.scores_.shape, dtype=bool)
        elif self.thresh == 0:
            return np.zeros(self.scores_.shape, dtype=bool)
        else:
            scores = _clean_nans(self.scores_)
            mask = np.zeros(scores.shape, dtype=bool)

            # Request a stable sort. Mergesort takes more memory (~40MB per
            # megafeature on x86-64).
            return np.array(scores > self.thresh, dtype=bool)



"""
Utility functions
"""
import os
from os import path
import logging
import numpy as np
import pandas as pd
from tqdm import tqdm
from functools import partial
from multiprocessing import Pool, cpu_count
from argparse import ArgumentParser

logger = logging.getLogger("IMG-EDIT-LEARN")
logging.basicConfig(level=logging.INFO)


def num_cores():
    n_cores = cpu_count()
    return max(1, n_cores - 1)


def imap_unordered_bar(func, args, n_proc=2):
    p = Pool(n_proc)
    ret = []
    with tqdm(total=len(args)) as pbar:
        for i, res in tqdm(enumerate(p.imap_unordered(func, args))):
            pbar.update()
            ret.append(res)
    pbar.close()
    p.close()
    p.join()
    return ret


def get_lines(fn):
    """get_lines

    :param fn:
    """
    if not os.path.exists(fn) or not os.path.isfile(fn):
        raise IOError("Invalid path given: {}".format(fn))
    with open(fn) as fp:
        lines = fp.read().splitlines()
    return lines


def file_parts(fn):
    base_name, ext = path.splitext(path.basename(fn))
    return path.dirname(fn), base_name, ext


def base_fn_add(fn, tok):
    base_dir, base_name, ext = file_parts(fn)
    return "{}{}{}{}".format(base_dir, base_name, tok, ext)


def fn_has_ext(ext_query, fn):
    _, _, ext = file_parts(fn)
    return ext[1:].lower() == ext_query.lower()


def raise_after_logging(exc, msg):
    logger.error(msg)
    raise exc(msg)


def remove_extension(fn):
    """remove_extension

    :param fn:
    """
    return os.path.splitext(fn)[0]


def xmp_nef_pairs(input_fn):
    # parse the passed file lists
    fns = sorted(get_lines(input_fn))

    # parse out the xmp and nef files
    xmp_fns = {
        path.splitext(fn)[0]: fn
        for fn in filter(partial(fn_has_ext, "xmp"), fns)
    }
    nef_fns = {
        path.splitext(fn)[0]: fn
        for fn in filter(partial(fn_has_ext, "nef"), fns)
    }

    # detect issues with the numbers of XMP and NEF files
    n_xmp = len(xmp_fns)
    n_nef = len(nef_fns)
    if n_xmp == 0:
        raise_after_logging(IOError, "No XMP files were found.")
    if n_nef == 0:
        raise_after_logging(IOError, "No NEF files were found.")
    if n_xmp != n_nef:
        logger.warning("A different number of XMP and NEF files were parsed. "
                       "# XMP: {}, # NEF: {}. "
                       "Only detected {{XMP, NEF}} pairs will be used.".format(n_xmp, n_nef))

    nef_fns_final = []
    xmp_fns_final = []
    for base_name, xmp_fn in xmp_fns.items():
        nef_fn = nef_fns.get(base_name, None)
        if nef_fn is not None:
            nef_fns_final.append(nef_fn)
            xmp_fns_final.append(xmp_fn)

    logger.info("Detected {} pairs of {{XMP, NEF}} files.".format(len(xmp_fns_final)))
    return xmp_fns_final, nef_fns_final


def mask_rejected_photos(xmp_df):
    # Adobe Lightroom does not store information in the xmp file on whether the photo was flagged as 'rejected',
    # so here it is assumed that a value of < -3 in the crs:Exposure2012 field or a value less than -99.0 in the
    # crs:Saturation field indicates 'rejected'
    return np.array(
        (xmp_df["crs:Exposure2012"] == -3) | (xmp_df["crs:Saturation"] < -99.0),
        dtype=bool
    )



"""
A collection of utility functions for file i/o.
"""
import os
from abc import ABCMeta

import pandas as pd


def extension(fn):
    """extension

    :param fn:
    """
    return os.path.splitext(fn)[1]


class Factory(object):
    """
    Abstract factory class to dynamically return a specific internal instance.
    Subclasses of this class should define the internal classes supported.
    """
    __metaclass__ = ABCMeta

    # defined by subclasses
    _classes = {}

    @staticmethod
    def __new__(cls, *args, **kwargs):
        """
        Return an instantiated instance of the correct type.
        :param name:
            Name of the class to create.
        :param args:
            Positional arguments to pass to the class constructor.
        :param kwargs:
            Keyword arguments to pass to the class constructor.
        :return:
            Initialized instance.
        """
        # check the mapping from name to class pointer.
        class_name = args[0]
        algo = cls._classes.get(class_name)

        # if a bad key is given, raise an error.
        if not algo:
            raise ValueError("The requested class {} does not exist."
                             "Supported classes are {}".format(class_name, cls._classes.keys()))

        # if all is well, initialize the instance and return it.
        return algo(*args[1:], **kwargs)


class BatchFileHandler(object):

    DEFAULT_BATCH_SIZE = 5000
    JOIN_FUNCS = {
        ".pkl": pd.concat,
        ".csv": pd.concat
    }
    SUPPORTED_EXTENSIONS = (".pkl", ".csv")

    def __init__(self, file_ext, batch_size):
        self.batch_size = batch_size or self.DEFAULT_BATCH_SIZE
        if file_ext not in self.SUPPORTED_EXTENSIONS:
            raise ValueError("Invalid file type specified: {}".format(file_ext))
        self.file_ext = file_ext
        self.join_func = BatchFileHandler.JOIN_FUNCS[self.file_ext]

    def verify_path(self, fn):
        """verify_path

        :param fn:
        """
        if not os.path.isfile(fn) or not os.path.exists(fn):
            raise IOError("Invalid path given for read: {}".format(fn))

    def __enter__(self):
        """__enter__"""
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """__exit__

        :param exc_type:
        :param exc_value:
        :param exc_tb:
        """
        if exc_value:
            raise exc_value


class BatchFileReader(BatchFileHandler):
    """
    Reads files into a DataFrame in a batched fashion.
    """
    READ_FUNCS = {
        ".pkl": pd.read_pickle,
        ".csv": pd.read_csv
    }

    def __init__(self, fns, batch_size=None):
        super(BatchFileReader, self).__init__(extension(fns[0]), batch_size)
        for fn in fns:
            self.verify_path(fn)
        self.fns = fns
        self.read_func = self.READ_FUNCS[self.file_ext]

    def read(self):
        """read"""
        return self.join_func([
            self.read_func(fn)
            for fn in self.fns
        ])


class BatchFileWriter(BatchFileHandler):
    """
    Writes a DataFrame to files in a batched fashion.
    """
    WRITE_FUNCS = {
        ".pkl": pd.DataFrame.to_pickle,
        ".csv": pd.DataFrame.to_csv
    }

    def __init__(self, fn, batch_size=None):
        """__init__

        :param df:
        :param fn:
        :param batch_size:
        """
        super(BatchFileWriter, self).__init__(extension(fn), batch_size)
        self.fn = fn
        self.write_func = self.WRITE_FUNCS[self.file_ext]

    def write(self, df):
        """write

        :param df:
        """
        self.write_func(df, self.fn)


class BatchFileIO(Factory):
    """
    Factory to return a batch reader/writer.
    """
    _classes = {
        "r": BatchFileReader,
        "w": BatchFileWriter
    }


def demo():
    """demo"""
    # batch read
    from glob import glob
    data = BatchFileIO('r', glob("*.pkl")).read()
    # print data.shape

    # batch write
    BatchFileIO('w', 'new_file.pkl').write(data)
    # print os.path.exists('new_file.pkl')


if __name__ == '__main__':
    demo()



import logging
import pickle
import numpy as np
import pandas as pd
from argparse import ArgumentParser
from os import path
from sklearn.externals import joblib
from ielearn.extract import extract
from ielearn.extract.xmp import update_xmp_files
from ielearn.predict.train import nmi, UniformPredictor
from ielearn.util import (
    xmp_nef_pairs,
    mask_rejected_photos
)

logger = logging.getLogger("IMG-EDIT-LEARN")
DIR = path.dirname(__file__)
MODELS_FN = path.join(DIR, "models.pkl")


def predict(xmp_fns, nef_fns, models_fn):
    # load the models from disk
    model_data = joblib.load(models_fn)
    feature_names = model_data["feature_names"]
    models = model_data["models"]

    # extract image embeddings, exif features, and xmp labels
    features, labels = extract.extract(xmp_fns, nef_fns)

    # identify out any features that weren't present for modeling
    unsupp_features = np.setdiff1d(features.columns, feature_names)
    if len(unsupp_features) > 0:
        logger.warning("Detected {} features that were not present for modeling"
                       " These values therefore cannot not be used during prediction: {}".format(
                           len(unsupp_features), unsupp_features))
        features.drop(unsupp_features, axis=1, inplace=True)

    # filter out photos marked as rejected in Adobe Lightroom
    rejected_mask = mask_rejected_photos(labels)
    n_rejected = int(rejected_mask.sum())
    if n_rejected > 0:
        logger.info("Filtering {}/{} files as rejected".format(n_rejected, len(rejected_mask)))
    labels = labels.loc[~rejected_mask]
    features = features.loc[~rejected_mask]

    # identify out any modeled features that are not present now (will be missing XMP data)
    missing_features = np.setdiff1d(feature_names, features.columns)
    if len(missing_features) > 0:
        logger.warning("Detected {} features not present in probe image data"
                       " These values will be imputed: {}".format(len(missing_features), missing_features))
        # fill in the missing data columns as nan in the correct order (set by the model_data dict).
        # these values will then be imputed by the individual model pipelines.
        nans = np.full_like(np.empty((len(features), len(missing_features))), np.nan)
        nan_df = pd.DataFrame(nans, columns=missing_features, index=features.index)
        features = pd.concat((features, nan_df), axis=1).loc[:, feature_names]

    # identify target XMP settings that don't have models
    label_keys = labels.keys()
    unmodeled_xmp = np.setdiff1d(list(label_keys), list(models.keys()))
    if len(unmodeled_xmp):
        logger.warning("No predictor available for {}/{} XMP settings."
                       " These values will not be changed: {}".format(
                           len(unmodeled_xmp), len(label_keys), unmodeled_xmp))

    # use image embedding and xmp data to predict xmp edits
    new_values = pd.DataFrame(np.nan, index=labels.index, columns=labels.columns)
    for label in labels.columns:
        if label in models:
            prediction = models[label]["predictor"].predict(features)

            # if an encoder was used with this model, invert the encodind
            encoder = models[label].get('encoder')
            if encoder:
                prediction = encoder.inverse_transform(prediction)
            new_values.loc[:, label] = prediction

    return new_values


def parse_args():
    """parse_args"""
    parser = ArgumentParser()
    parser.add_argument(dest="input_fn",
                        help="Path to a file which contains a list of NEF and XMP files to parse (one per line).")
    parser.add_argument('--models_fn', dest='models_fn', default=MODELS_FN, required=False,
                        help='An existing models pickle file to use.')
    return parser.parse_args()


def cli():
    if __name__ == "__main__":
        args = parse_args()
        xmp_fns, nef_fns = xmp_nef_pairs(args.input_fn)
        update_xmp_files(predict(xmp_fns, nef_fns, args.models_fn))
cli()



"""
TODO description
"""
import os
import re
import pickle
from os import path
import logging
from argparse import ArgumentParser
from collections import Counter, defaultdict

import numpy as np
import pandas as pd
import seaborn as sns
# from matplotlib import pyplot as plt
from sklearn.externals import joblib
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import RobustScaler, LabelEncoder, Imputer
from sklearn.ensemble import (
    # RandomForestRegressor,
    #RandomForestClassifier,
    GradientBoostingRegressor,
    GradientBoostingClassifier,
    # IsolationForest
)
from sklearn.mixture import GaussianMixture as GMM
from sklearn.cluster import KMeans
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.utils.validation import check_array, check_X_y
from sklearn.model_selection import ShuffleSplit, train_test_split, GridSearchCV
from sklearn.metrics import r2_score, mean_squared_error, f1_score, accuracy_score, make_scorer
from sklearn.feature_selection import mutual_info_regression

from ielearn.extract.xmp import parse_target_types
from ielearn.predict.feature_selection import SelectThreshold

sns.set(color_codes=True)
logger = logging.getLogger("IMG-EDIT-LEARN")
logging.basicConfig(level=logging.INFO)

MODELS_FN = "models.pkl"


class UniformPredictor(BaseEstimator, ClassifierMixin):

    def __init__(self, uniform_prediction=0, **kwargs):
        self.uniform_prediction = uniform_prediction

    def fit(self, X, y):
        X, y = check_X_y(X, y)
        self.uniform_prediction = pd.value_counts(y).idxmax()
        return self

    def predict(self, X):
        X = check_array(X)
        return self.uniform_prediction * np.ones((X.shape[0], 1))


FN_DESIRED_FIELDS = os.path.join("..", "extract", "res", "desired_fields")
SPECIAL_CHARS = re.compile("\W+")
CLF_AND_SCORER = {
    "categorical": (GradientBoostingClassifier(), "Acc", accuracy_score),
    "categorical_uniform": (UniformPredictor(), "Acc", accuracy_score),
    "numerical": (GradientBoostingRegressor(), "MSE", mean_squared_error),
    "numerical_uniform": (UniformPredictor(), "MSE", mean_squared_error)
}



def select_features(df, target_types):
    """select_features

    :param df:
    """
    logger.info("{} samples".format(df.shape[0]))
    embedding_cols = ["F{}".format(i + 1) for i in range(512)]
    exif_cols = [col for col in df.columns if "exif" in col]
    feature_cols = embedding_cols + exif_cols
    features = df.loc[:, feature_cols]

    logger.info("{} embedding features".format(len(embedding_cols)))
    logger.info("{} exif data features".format(len(exif_cols)))
    logger.info("{} features".format(features.shape[1]))

    # transform exif features using one-hot or standardization
    # for col in exif_cols:
    #     target_type = target_types[col]
    #     if target_type == "binary":
    #         new_features = binarizer.fit(features[col])

    #null_mask = features.isnull().sum() > 0
    #logger.info("Removing {} features with more than 0 null values: {}".format(
    #    len(null_mask[null_mask]),
    #    features.columns[null_mask]))
    #features = features.loc[:, ~null_mask]

    logger.info("Finalized {} features.".format(features.shape[1]))
    return features


def detect_label_outliers(y, encoder=None):
    # remove outliers
    # TODO per-class outlier removal also

    # if an encoder is provided, use it to invert the encoding before computing stats.
    if encoder:
        y = encoder.inverse_transform(y)

    # perform outlier removal on mean/var-normalized samples (after filling missing data).
    y_std = RobustScaler().fit_transform(
        Imputer(strategy="median").fit_transform(
            y.reshape([-1, 1])
        )
    ).reshape([-1])
    mask = (y_std < y_std.mean() - 3*y_std.std()) | (y_std > y_std.mean() + 3*y_std.std())
    logger.info("Filtering {}/{} as outliers.".format(mask.sum(), len(mask)))
    return mask


def select_labels(df, target_types):
    """select_labels

    :param df:
    :param target_types:
    """

    def to_lookup(tok):
        """to_lookup

        :param tok:
        """
        pos = tok.find("_")
        if pos > -1:
            return tok[:pos]
        return tok

    label_cols = [col for col in df.columns if "crs" in col]
    labels = df.loc[:, label_cols]
    logger.info("Started with {} labels".format(labels.shape[1]))

    no_tone_curve_cols = [col for col in labels.columns if "ToneCurve" not in col]
    logger.info("\nRemoving {} tone curve labels.".format(len(no_tone_curve_cols)))
    labels = labels.loc[:, labels.columns.isin(no_tone_curve_cols)]

    # note that these are missing labels, not missing features. so we need semi-supervised
    # learning, not data imputation.
    # TODO make sure this is ok
    num_missing_labels = labels.isnull().sum()
    incomplete_label_counts = dict(num_missing_labels[num_missing_labels > 0])
    if len(incomplete_label_counts) > 0:
        logger.warning("Missing labels")
        for label, n_missing in incomplete_label_counts.items():
            logger.warning("  - {label}: {num} missing.".format(label=label, num=n_missing))
            to_impute = labels[label].values.reshape((-1, 1))
            label_imputer = Imputer(strategy="most_frequent").fit(to_impute)
            labels.loc[:, label] = label_imputer.transform(to_impute)
        logger.error("Zero-filling potentially a ton of labels. Verify after prototyping.")

    # remove labels with negligible variance
    MIN_VAR = 0.001
    low_var = RobustScaler().fit_transform(labels).var(axis=0) < MIN_VAR
    logger.info("Removing {}/{} labels with small variance.".format(low_var.sum(), len(low_var)))
    logger.info(labels.columns[low_var])
    labels = labels.loc[:, ~low_var]

    # converting regression -> classification for targets with a small number of unique values
    CATEG_NUMER_THRESH = 10
    # maximum prior of any class for a target label to be selected for modeling. above this threshold, you're
    # better off predicting that class.
    MAX_SINGLE_CLASS_PRIOR = 0.99
    # minimum class prior for the class to remain intact (and not be ignored)
    MIN_SINGLE_CLASS_PRIOR = 0.001
    n_convert = 0
    n_dominant_prior = 0
    labels_data = defaultdict(lambda: dict())
    new_categ_classes = []
    for col in labels.columns:
        # convert a numberical label to categorical if there are only a "few" unique values
        # if the resulting categorical distribution is dominated by one class, don't perform any modeling
        n_unique = len(labels[col].unique())
        few_n_unique = n_unique < CATEG_NUMER_THRESH

        label_counts = labels[col].value_counts().sort_values(ascending=False)
        label_counts /= label_counts.sum()
        if label_counts.values[0] >= MAX_SINGLE_CLASS_PRIOR:
            logger.warning("Target label {} will not be modeled, as its main class {} has a prior of {:.5f}, which "
                           "is higher than the maximum allowed of {}.".format(
                               col, label_counts.index[0], label_counts.values[0], MAX_SINGLE_CLASS_PRIOR))
            # TODO this needs to turn into a uniform predictor, not avoiding modeling altogether.
            labels_data[col]["data"] = np.full_like(y, label_counts.index[0])
            labels_data[col]["type"] = "categorical_uniform"
            n_dominant_prior += 1
        else:
            # TODO don't just use a threshold number of unique values, choose based on the modes of the distribution
            # TODO generalize Temperature model VQ to other models
            y = labels[col].values
            if col == "crs:Temperature":
                # TODO Quadratic Discriminant Analysis

                outlier_mask = detect_label_outliers(y)
                y = y[~outlier_mask]

                #clus = KMeans(n_clusters=5).fit(y.reshape([-1, 1]))
                #modes = clus.cluster_centers_
                clus = GMM(
                    n_components=5,
                    covariance_type='diag',
                    reg_covar=1e-3,
                    means_init=np.array([7200, 6430, 5177, 3774, 2500]).reshape([-1, 1])
                ).fit(y.reshape([-1, 1]))
                modes = clus.means_

                distances = np.abs(modes.reshape([1, -1]) - y.reshape([-1, 1]))
                inds = np.argmin(distances, axis=1)
                y_vq = modes.astype(int)[inds].reshape([-1,])
                logger.info("Fitted Temperature Modes: {}".format(modes))
                encoder = LabelEncoder().fit(y_vq)
                labels_data[col]["data"] = encoder.transform(y_vq)
                labels_data[col]["type"] = "categorical"
                labels_data[col]["encoder"] = encoder
                labels_data[col]["sample_mask"] = ~outlier_mask
                n_convert += 1
                new_categ_classes.append(col)
            else:
                if few_n_unique:
                    classes_insuff_prior = label_counts.index[label_counts.values < MIN_SINGLE_CLASS_PRIOR]
                    use_mask = ~np.isin(y, classes_insuff_prior)
                    labels_data[col]["sample_mask"] = use_mask
                    y = y[use_mask]
                    if len(classes_insuff_prior) > 0:
                        logger.info("Targets {} ({}/{}) have priors under the configured minimum of {}. This led to a "
                                    "removal of {}/{} samples.".format(
                                        classes_insuff_prior, len(classes_insuff_prior), len(label_counts),
                                        MIN_SINGLE_CLASS_PRIOR, np.sum(~use_mask), len(use_mask)))

                    n_convert += 1
                    new_categ_classes.append(col)
                    labels_data[col]["type"] = "categorical"
                    logger.info("Class {} converted to categorical with {} target classes.".format(col, n_unique))
                else:
                    labels_data[col]["type"] = target_types.loc[to_lookup(col), 'prediction_type']

                # for categorical labels, use a label encoder and store the encoder. either way, store the data
                if labels_data[col]["type"] == "categorical":
                    encoder = LabelEncoder().fit(labels[col])
                    labels_data[col]["data"] = encoder.transform(y)
                    labels_data[col]["encoder"] = encoder
                else:
                    labels_data[col]["data"] = y

    msg = "{}/{} labels were not modeled due to a dominant target class or value.\n".format(
        n_dominant_prior, len(labels.columns))
    msg += "Converted {}/{} labels to categorical.\n".format(n_convert, len(labels.columns))
    msg += "Finalized {} labels.\n".format(len(labels_data))
    logger.info(msg)
    logger.info((Counter(
        [d["type"] for d in list(labels_data.values())]
    )))

    return labels_data


def nmi(X, y):
    """
    Normalized mutual information between X and y.
    :param X:
    :param y:
    """
    mi = mutual_info_regression(X, y)
    return mi / mi.max()


def build_uniform_predictor(label, X, model_info, inner_clf, metric_name, metric_func):
    y, target_type = model_info["data"], model_info["type"]
    sample_mask = model_info.get('sample_mask', np.ones(len(y), dtype=bool))
    X = X.loc[sample_mask]
    clf = Pipeline([
        ("impute", Imputer(strategy="most_frequent" if target_type == "categorical" else "mean")),
        ("predict", inner_clf)
    ])
    clf.fit(X, y)
    return clf, None


def get_train_test_split(X, y, is_categorical=False, log_counts=False):
    """

    :param X:
    :param y:
    """
    label_counts = pd.Series(Counter(y)).sort_values(ascending=False)
    if is_categorical and label_counts.values[1:].sum() > 5:
        # we want to stratify the labels, but minority classes will likely only contain a few samples.
        # so we'll stratify by the top few labels only.
        logger.info("Stratifying the train-test split.")
        label_ratios = (label_counts / label_counts.sum())
        top_labels = label_ratios.index[:1]
        y_strat = pd.Series(y).isin(top_labels).values

        stratified_splitter = ShuffleSplit(n_splits=1, train_size=0.8)
        train_ind, test_ind = next(stratified_splitter.split(X, y_strat))
        X_train, X_test, y_train, y_test = X.iloc[train_ind], X.iloc[test_ind], y[train_ind], y[test_ind]
    else:
        logger.info("Not stratifying the train-test split.")
        splitter = ShuffleSplit(n_splits=1, train_size=0.8)
        train_ind, test_ind = next(splitter.split(X, y))
        X_train, y_train, X_test, y_test = X.iloc[train_ind], y[train_ind], X.iloc[test_ind], y[test_ind]

    logger.info("X_train: {}".format(X_train.shape))
    logger.info("X_test: {}".format(X_test.shape))
    logger.info("y_train: {}".format(y_train.shape))
    logger.info("y_test: {}".format(y_test.shape))
    if is_categorical:
        logger.info("y class counts: {}".format(Counter(y)))
        logger.info("y_train class counts: {}".format(Counter(y_train)))
        logger.info("y_test class counts: {}".format(Counter(y_test)))

    return X_train, y_train, X_test, y_test


def build_predictor(label, X, model_info, inner_clf, metric_name, metric_func):
    """build_model

    :param X:
    :param y:
    :param target_type:
    """
    y, target_type, encoder = model_info["data"], model_info["type"], model_info.get("encoder")
    sample_mask = model_info.get('sample_mask', np.ones(len(y), dtype=bool))
    logger.info("Sample mask: {}".format(Counter(sample_mask)))
    X = X.loc[sample_mask]

    # if the majority class isn't totally dominant, stratify the train-test split.
    X_train, y_train, X_test, y_test = get_train_test_split(X, y, is_categorical="categorical" in target_type)

    # parameter fitting
    params = {
        "n_estimators": [250],
        "learning_rate": [1e-2, 0.05],
        "max_depth": [3, 4],
        "subsample": [0.9],
        "min_samples_split": [2, 1e-2],
        "min_samples_leaf": [1, 1e-2],
        # "max_features": [None, 'auto'],
    }
    clf = Pipeline([
        ("impute", Imputer(strategy="most_frequent" if target_type == "categorical" else "mean")),
        ("select", SelectThreshold(nmi, thresh=0.2)),
        ("standardize", RobustScaler()),
        ("predict", GridSearchCV(inner_clf, params,
                                 scoring=make_scorer(metric_func), n_jobs=-1, verbose=0, return_train_score=True))
    ])
    grid_search = clf.named_steps["predict"]
    clf.fit(X_train, y_train)

    # performance on the training data
    y_pred_train = clf.predict(X_train)
    logger.info(("({label}) Performance on Train ({metric_name}): {metric_value}".format(
        label=label, metric_name=metric_name, metric_value=metric_func(y_train, y_pred_train))
    ))

    # performance on the evaluation data
    y_pred = clf.predict(X_test)
    logger.info(("({label}) Performance on Test ({metric_name}): {metric_value}".format(
        label=label, metric_name=metric_name, metric_value=metric_func(y_test, y_pred))
    ))

    # grab some desired data from the grid search to store to disk
    desired_columns = ["mean_test_score", "mean_train_score"] +\
        list([c for c in list(grid_search.cv_results_.keys()) if "param_" in c]) +\
        list([c for c in list(grid_search.cv_results_.keys()) if "rank" in c])
    results = pd.DataFrame.from_dict(grid_search.cv_results_).loc[:, desired_columns].sort_values("rank_test_score")
    return clf, results


def build_model(label, X, model_info, inner_clf, metric_name, metric_func):
    """build_model

    :param X:
    :param y:
    :param target_type:
    """
    FUNC_MAPPING = {
        "categorical": build_predictor,
        "categorical_uniform": build_uniform_predictor,
        "numerical": build_predictor,
        "numerical_uniform": build_uniform_predictor
    }

    y, target_type, encoder = model_info["data"], model_info["type"], model_info.get("encoder")
    func = FUNC_MAPPING[target_type]
    return func(label, X, model_info, inner_clf, metric_name, metric_func)


def ensure_dir(dir_name):
    if not path.exists(dir_name):
        os.makedirs(dir_name)


def train_models(features_fn, labels_fn, target_label):
    features_df = pd.read_csv(features_fn)
    labels_df = pd.read_csv(labels_fn)
    target_types = parse_target_types().replace("binary", "categorical")

    # prep features
    features = select_features(features_df, target_types)

    # prep labels
    labels_data = select_labels(labels_df, target_types)

    # save a list of the feature names for prediction-time
    data = {
        "feature_names": features.columns,
        "models": {}
    }

    for label, train_info in labels_data.items():
        if target_label is None or label == target_label:
            logger.info("Training {model_type} model for {label}".format(model_type=train_info["type"], label=label))

            target_type = train_info["type"]
            inner_clf, metric_name, metric_func = CLF_AND_SCORER[target_type]
            clf, results = build_model(label, features, train_info, inner_clf, metric_name, metric_func)
            data["models"][label] = {
                "encoder": train_info.get("encoder"),
                "predictor": clf,
                "stats": results
            }

    return data


def persist_models(model_data, fn):
    # only if there are models to persist
    n_models = len(model_data["models"])
    if n_models > 0:
        # if the models file already exists, update the data structure with the passed models and re-write to disk.
        # else, persist the passed data structure.
        if path.exists(fn):
            logger.info("Updating {} with the {} new finalized models.".format(fn, n_models))
            existing_model_data = joblib.load(fn)
            for name, data in model_data["models"].items():
                existing_model_data['models'][name] = data
            joblib.dump(existing_model_data, fn)
        else:
            logger.info("Persisting the {} finalized models to {}.".format(n_models, fn))
            joblib.dump(model_data, fn)

    else:
        logger.warning("Not updating models in {} since no models were passed to persist.".format(fn))


def cli():
    if __name__ == '__main__':

        def parse_args():
            """parse_args
            """
            parser = ArgumentParser()
            parser.add_argument(dest='features_fn', help='Path to csv data with features.')
            parser.add_argument(dest='labels_fn', help='Path to csv data with labels.')
            parser.add_argument('--target_label', dest='target_label', default=None, required=False,
                                help='A single target label to model.')
            parser.add_argument('--models_fn', dest='models_fn', default=MODELS_FN, required=False,
                                help='An existing models pickle file to use.')
            args = parser.parse_args()
            return args

        args = parse_args()
        model_data = train_models(args.features_fn, args.labels_fn, args.target_label)
        persist_models(model_data, args.models_fn)

cli()



"""
Extract a data set from NEF and XMP files.
"""
from os import path
import logging
import numpy as np
import pandas as pd
from tqdm import tqdm
from functools import partial
from multiprocessing import Pool
from itertools import filterfalse
from argparse import ArgumentParser

from ielearn.util import (
    imap_unordered_bar,
    get_lines,
    fn_has_ext,
    xmp_nef_pairs,
    base_fn_add,
    raise_after_logging,
    remove_extension,
    mask_rejected_photos
)
from ielearn.extract import (
    embedding,
    xmp
)

logger = logging.getLogger("IMG-EDIT-LEARN")
logging.basicConfig(level=logging.INFO)


def is_exif_descriptor(d):
    return d.startswith("exif:")


def extract(xmp_fns, nef_fns):
    """extract"""
    if len(xmp_fns) != len(nef_fns):
        raise ValueError("# of XMP and NEF file names differ!")

    # parse the xmp files
    xmp_df = xmp.run_extraction(xmp_fns)

    # extract embeddings from the images
    embedding_df = embedding.run_extraction(nef_fns)

    # filter out photos marked as rejected in Adobe Lightroom
    rejected_mask = mask_rejected_photos(xmp_df)
    n_rejected = int(rejected_mask.sum())
    if n_rejected > 0:
        logger.info("Filtering {}/{} files as rejected".format(n_rejected, len(rejected_mask)))
    xmp_df = xmp_df.loc[~rejected_mask]
    embedding_df = embedding_df.loc[~rejected_mask]

    # segregate features and labels
    exif_cols = list(filter(is_exif_descriptor, xmp_df.columns))
    non_exif_cols = list(filterfalse(is_exif_descriptor, xmp_df.columns))
    features = pd.concat((
        embedding_df,
        xmp_df.loc[:, exif_cols]
    ), axis=1).set_index("fn")
    labels = xmp_df.loc[:, non_exif_cols].set_index("fn")
    return features, labels

    # merge the two DataFrames by their file name (with extension removed)
    # merge_col = 'fn_trunc'
    # xmp_df[merge_col] = xmp_df['fn'].map(remove_extension)
    # embedding_df[merge_col] = embedding_df['fn'].map(remove_extension)
    # del embedding_df['fn']
    # main_df = xmp_df.merge(embedding_df, how='inner', on=merge_col)
    # del main_df[merge_col]
    # return main_df


def parse_args():
    """parse_args"""
    parser = ArgumentParser()
    parser.add_argument(dest="input_fn",
                        help="Path to a file which contains a list of NEF and XMP files to parse (one per line).")
    parser.add_argument(dest="base_fn",
                        help="Base path to where the parsed data sets should be written to.")
    return parser.parse_args()


def cli():
    if __name__ == "__main__":
        args = parse_args()
        xmp_fns, nef_fns = xmp_nef_pairs(args.input_fn)
        features, labels = extract(xmp_fns, nef_fns)
        features.to_csv(base_fn_add(args.base_fn, ".features"))
        labels.to_csv(base_fn_add(args.base_fn, ".labels"))
cli()



"""
XMP File processing (analysis and synthesis).
"""
import os
import logging
import numpy as np
import pandas as pd
from tqdm import tqdm
from shutil import copyfile
from multiprocessing import cpu_count, Pool
from tempfile import NamedTemporaryFile
from functools import partial
from subprocess import check_call
from ielearn.util import (
    imap_unordered_bar,
    num_cores
)
from libxmp import XMPFiles
from libxmp.utils import file_to_dict
from libxmp.consts import (
    XMP_NS_EXIF_Aux,
    XMP_NS_Photoshop,
    XMP_NS_EXIF,
    XMP_NS_XMP,
    XMP_NS_DC,
    XMP_NS_XMP_MM,
    XMP_NS_CameraRaw,
    XMP_NS_TIFF
)

XMP_PROPERTIES = (
    XMP_NS_EXIF,
    XMP_NS_EXIF_Aux,
    XMP_NS_Photoshop,
    XMP_NS_XMP,
    XMP_NS_DC,
    XMP_NS_XMP_MM,
    XMP_NS_CameraRaw,
    XMP_NS_TIFF
)
PROPERTIES = (
    XMP_NS_EXIF,
    XMP_NS_TIFF
)
FN_TYPE_MAP = os.path.join(os.path.dirname(__file__), "res", "type_map.csv")
XMP_XPACKET_HEADER = "<?xpacket begin=\"\" id=\"W5M0MpCehiHzreSzNTczkc9d\"?>"
XMP_XPACKET_FOOTER = "<?xpacket end=\"w\"?>"
XMP_NS = {
    "crs": XMP_NS_CameraRaw,
    "exif": XMP_NS_EXIF
}

logger = logging.getLogger("IMG-EDIT-LEARN")
logging.basicConfig(level=logging.INFO)


def lightroom_compatible(name, value, type_map):

    # some properties expect an explicit sign symbol
    sign = type_map.loc[name, 'sign']
    prefix = ""
    if sign == "signed":
        prefix = "\+" if value > 0 else ""

    # parse based on the data type
    dtype = type_map.loc[name, 'dtype']
    if dtype == 'bool':
        if value == 1:
            rv = "True"
        elif value == 0:
            rv = "False"
        else:
            # fail-safe against a non-binary prediction
            logger.error("Non-binary output {} for expected-binary field {}".format(value, name))
    elif dtype == 'int':
        rv = "{}{}".format(prefix, int(value))
    elif dtype == 'float':
        return "{}{:.2f}".format(prefix, float(value))
    else:
        logger.error("Unplanned dtype {}.".format(dtype))
    return rv


def update_single_xmp_file(update_fn_vals, type_map):
    fn = update_fn_vals.name
    logger.info(fn)

    # TODO only for dev
    temp_fn = fn + ".bak"
    copyfile(fn, temp_fn)

    xmp_meta = file_to_meta(fn)
    for name, value in update_fn_vals.iteritems():
        cmd = "sed -i 's/{name}=\".*\"/{name}=\"{value}\"/g' \"{fn}\"".format(
            name=name,
            value=lightroom_compatible(name, value, type_map),
            fn=fn
        )
        check_call(cmd, shell=True)


def update_xmp_file_batch(update_vals, type_map):
    update_vals.apply(update_single_xmp_file, args=tuple([type_map]), axis=1)


def update_xmp_files(update_vals):
    # make updates to the XMP settings
    type_map = parse_target_types()
    update_vals.dropna(axis=1, inplace=True)
    logger.info("# files to update: {}".format(len(update_vals)))
    n_cores = num_cores()
    df_parts = np.array_split(update_vals, n_cores)
    with Pool(n_cores) as pool:
        pool.map(partial(update_xmp_file_batch, type_map=type_map), df_parts)


def file_to_meta(fn):
    xmp_files = XMPFiles()

    with open(fn, 'r') as fp:
        header = fp.readline()
    if "xpacket" in header:
        # the file is already in a compatible format for the XMP parser.
        xmp_files.open_file(fn, open_read=True)
        xmp_obj = xmp_files.get_xmp()
    else:
        # need to wrap the file with a header and footer that allows
        # the XMP parser to parse the file into a dict.
        # we will only transform the data in a temporary file, leaving
        # the original file untouched.
        with NamedTemporaryFile(mode='w', delete=False) as fp,\
                open(fn, 'r') as raw_fp:
            temp_fn = fp.name
            fp.write(XMP_XPACKET_HEADER + "\n")
            for line in raw_fp:
                fp.write("{line}\n".format(line=line))
            fp.write(XMP_XPACKET_FOOTER + "\n")
        xmp_files.open_file(temp_fn, open_read=True, open_onlyxmp=True)
        xmp_obj = xmp_files.get_xmp()
        os.remove(temp_fn)

    return xmp_obj


# def update_xmp_file(fn, xmp_obj):
#     # make changes here
#     xmp_obj.get_property(XMP_NS_CameraRaw, "RawFileName")
#     xmp_obj.set_property(XMP_NS_CameraRaw, "RawFileName", "hey")
#     xmp_obj.get_property(XMP_NS_CameraRaw, "RawFileName")
#     xmp_obj.set_array_item(XMP_NS_CameraRaw, "ToneCurve", 1, "99, 99")
#
#     # dump the updated object to the file
#     with open(fn, 'w') as fp:
#         fp.write(str(xmp_obj))


# def file_to_meta(fn):
#     with open(fn, 'r') as fp:
#         header = fp.readline()
#     if "xpacket" in header:
#         # the file is already in a compatible format for the XMP parser.
#         return file_to_dict(fn)
#
#     # need to wrap the file with a header and footer that allows
#     # the XMP parser to parse the file into a dict.
#     # we will only transform the data in a temporary file, leaving
#     # the original file untouched.
#     with NamedTemporaryFile(mode='w', delete=False) as fp,\
#             open(fn, 'r') as raw_fp:
#         temp_fn = fp.name
#         fp.write(XMP_XPACKET_HEADER + "\n")
#         for line in raw_fp:
#             fp.write("{line}\n".format(line=line))
#         fp.write(XMP_XPACKET_FOOTER + "\n")
#
#     xmp_files = XMPFiles()
#     xmp_files.open_file(temp_fn, open_read=True, open_onlyxmp=True)
#     xmp_obj = xmp_files.get_xmp()
#     os.remove(temp_fn)
#     return xmp_obj


def parse_target_types():
    return pd.read_csv(
        FN_TYPE_MAP,
        index_col=0,
        names=['prediction_type', 'sign', 'dtype']
    )


def parse_floats(s):
    """parse_floats

    :param s:
    """
    if isinstance(s, str):
        if "/" in s:
            # parse a ratio to its float value
            num, den = s.split("/")
            return [float(num) / float(den)]
        elif "," in s:
            # parse a csv variable into multiple new columns
            return [float(el) for el in s.split(",")]
        else:
            # parse to float directly
            return [float(s)]
    else:
        # parse to float directly
        return [float(s)]


def convert_types(df, type_map):
    """convert_types

    :param df:
    """

    data = []
    data_fields = []
    logging.info("Converting data types for the parsed XMP data.")
    for column in tqdm(df.columns):
        dtype = type_map.loc[column, 'prediction_type']
        if not dtype:
            raise TypeError("Unexpected type {} for property {}.".format(dtype, column))

        if dtype == "categorical":
            values = pd.get_dummies(df[column]).values.tolist()
            data.extend(list(zip(*values)))
            data_fields.extend(["{}_{}".format(column, i) for i in range(len(values[0]))])
        elif dtype == "binary":
            data.append(df[column].fillna(0).replace({"True": 1, "False": 0}).astype(int).values.tolist())
            data_fields.append(column)
        else:
            # dtype == "numerical"
            values = df[column].replace('', np.nan).apply(parse_floats).values.tolist()
            lengths = np.array([len(val) if isinstance(val, list) else 1 for val in values])
            target_len = np.max(lengths)
            if target_len > 1:
                for i, val in enumerate(values):
                    if lengths[i] < target_len:
                        values[i] = [None] * target_len
                data_fields.extend(["{}_{}".format(column, i) for i in range(target_len)])
            else:
                data_fields.append(column)

            data.extend(list(zip(*values)))

        if len(data) != len(data_fields):
            raise RuntimeError("The number of data data_fields and the number of data column names is different.")

    return data_fields, data


def parse_xmp_data(fn):
    with open(fn, 'r') as fp:
        header = fp.readline()
    if "xpacket" in header:
        # the file is already in a compatible format for the XMP parser.
        return file_to_dict(fn)

    # need to wrap the file with a header and footer that allows
    # the XMP parser to parse the file into a dict.
    # we will only transform the data in a temporary file, leaving
    # the original file untouched.
    with NamedTemporaryFile(mode='w', delete=False) as fp,\
            open(fn, 'r') as raw_fp:
        temp_fn = fp.name
        fp.write(XMP_XPACKET_HEADER + "\n")
        for line in raw_fp:
            fp.write("{line}\n".format(line=line))
        fp.write(XMP_XPACKET_FOOTER + "\n")
    xmp_data = file_to_dict(temp_fn)
    os.remove(temp_fn)
    return xmp_data


def xmp_to_vec(fn, type_map):
    """xmp_to_vec

    :param fn:
    :param type_map:
    """
    # read in the core data of interest from the XMP file.
    xmp_data = parse_xmp_data(fn)
    df = pd.DataFrame(
        [
            tup[:2]
            for _, data in list(xmp_data.items())
            for tup in data
        ],
        columns=["field", "value"]
    )

    # filter down to the desired properties only.
    df = df.loc[df['field'].isin(type_map.index)]

    # return a mapping from the desired properties to their values.
    return {
        field: value
        for field, value in zip(df["field"].values, df["value"].values)
    }


def xmp_extract(fns, type_map):
    """xmp_extract

    :param fns:
    :param type_map:
    """
    logger.info("Extracting raw XMP data.")
    func = partial(xmp_to_vec, type_map=type_map)
    xmp_to_vec(fns[0], type_map=type_map)
    xmp_data = imap_unordered_bar(func, fns, n_proc=2)
    xmp_data = pd.DataFrame(xmp_data)

    # convert the data types
    data_fields, data = convert_types(xmp_data, type_map)
    df = pd.DataFrame(data).transpose()
    df.columns = data_fields
    df['fn'] = fns

    return df


def run_extraction(fns):
    """run_extraction

    :param fns:
    """
    logger.info("Extracting XMP and EXIF data from {} XMP data files.".format(len(fns)))

    # parse the map of prediction target dtypes
    type_map = parse_target_types()

    # parse the labels from each xmp file
    df = xmp_extract(fns, type_map)

    return df



"""
Extract neural embeddings for all of the files specified in the passed file name list.
"""
import os
import pickle
import logging
from tqdm import tqdm
import multiprocessing as mp

import numpy as np
import pandas as pd
from numpy.linalg import norm
from keras.applications.vgg16 import VGG16
from keras.preprocessing.image import load_img, img_to_array
from keras.applications.vgg16 import preprocess_input

logger = logging.getLogger("IMG-EDIT-LEARN")
logging.basicConfig(level=logging.INFO)

INPUT_SHAPE = (224, 224, 3)
OUTPUT_DIMENSION = 512
MODEL = VGG16(
    weights='imagenet',
    input_shape=(INPUT_SHAPE[0], INPUT_SHAPE[1], INPUT_SHAPE[2]), pooling='max',
    include_top=False
)


def extract_embedding(img_path):
    """
    Use a pre-trained model to extract the embedding of the picture located at img_path.
    """
    img = img_to_array(load_img(img_path, target_size=(INPUT_SHAPE[0], INPUT_SHAPE[1])))
    img = preprocess_input(np.expand_dims(img, axis=0))
    feat = MODEL.predict(img)
    norm_feat = feat[0] / norm(feat[0])
    return norm_feat


def run_extraction(fns):
    """run_extraction
    Extract features and index the images.

    :param fns:
    """
    logger.info("Extracting neural embeddings from {} NEF images.".format(len(fns)))

    df = pd.DataFrame(
        [extract_embedding(fn) for fn in tqdm(fns)],
        columns=["F{}".format(i+1) for i in range(OUTPUT_DIMENSION)]
    )
    df['fn'] = fns
    return df
